@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2012-2015] Novell, Inc.
- * Copyright (c) [2016,2018] SUSE LLC
+ * Copyright (c) [2016-2024] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -28,16 +28,18 @@
 #include <string>
 #include <list>
 #include <queue>
-#include <set>
+#include <map>
 #include <boost/thread.hpp>
 
 #include <snapper/Snapper.h>
 #include <snapper/Snapshot.h>
 #include <snapper/Comparison.h>
+#include <snapper/Plugins.h>
 #include <dbus/DBusConnection.h>
 #include <dbus/DBusMessage.h>
 
 #include "MetaSnapper.h"
+#include "FilesTransferTask.h"
 
 
 using namespace std;
@@ -82,7 +84,7 @@ public:
     void signal_snapshot_modified(DBus::Connection& conn, const string& config_name,
 				  unsigned int num);
     void signal_snapshots_deleted(DBus::Connection& conn, const string& config_name,
-				  const list<dbus_uint32_t>& nums);
+				  const vector<dbus_uint32_t>& nums);
 
     void list_configs(DBus::Connection& conn, DBus::Message& msg);
     void get_config(DBus::Connection& conn, DBus::Message& msg);
@@ -101,6 +103,10 @@ public:
     void create_pre_snapshot(DBus::Connection& conn, DBus::Message& msg);
     void create_post_snapshot(DBus::Connection& conn, DBus::Message& msg);
     void delete_snapshots(DBus::Connection& conn, DBus::Message& msg);
+    void is_snapshot_read_only(DBus::Connection& conn, DBus::Message& msg);
+    void set_snapshot_read_only(DBus::Connection& conn, DBus::Message& msg);
+    void get_default_snapshot(DBus::Connection& conn, DBus::Message& msg);
+    void get_active_snapshot(DBus::Connection& conn, DBus::Message& msg);
     void calculate_used_space(DBus::Connection& conn, DBus::Message& msg);
     void get_used_space(DBus::Connection& conn, DBus::Message& msg);
     void mount_snapshot(DBus::Connection& conn, DBus::Message& msg);
@@ -109,15 +115,19 @@ public:
     void create_comparison(DBus::Connection& conn, DBus::Message& msg);
     void delete_comparison(DBus::Connection& conn, DBus::Message& msg);
     void get_files(DBus::Connection& conn, DBus::Message& msg);
+    void get_files_by_pipe(DBus::Connection& conn, DBus::Message& msg);
     void setup_quota(DBus::Connection& conn, DBus::Message& msg);
     void prepare_quota(DBus::Connection& conn, DBus::Message& msg);
     void query_quota(DBus::Connection& conn, DBus::Message& msg);
+    void query_free_space(DBus::Connection& conn, DBus::Message& msg);
     void sync(DBus::Connection& conn, DBus::Message& msg);
-    void debug(DBus::Connection& conn, DBus::Message& msg) const;
+    void get_plugins_report(DBus::Connection& conn, DBus::Message& msg);
+    void clear_plugins_report(DBus::Connection& conn, DBus::Message& msg);
+    void debug(DBus::Connection& conn, DBus::Message& msg);
 
     void dispatch(DBus::Connection& conn, DBus::Message& msg);
 
-    Client(const string& name, const Clients& clients);
+    Client(const string& name, uid_t uid, const Clients& clients);
     ~Client();
 
     list<Comparison*>::iterator find_comparison(Snapper* snapper, unsigned int number1,
@@ -137,33 +147,42 @@ public:
     void remove_mount(const string& config_name, unsigned int number);
 
     const string name;
+    const uid_t uid;
 
     list<Comparison*> comparisons;
 
-    set<string> locks;
+    map<string, unsigned int> locks;
 
     map<pair<string, unsigned int>, unsigned int> mounts;
 
-    struct Task
+    struct MethodCallTask
     {
-	Task(DBus::Connection& conn, DBus::Message& msg) : conn(conn), msg(msg) {}
+	MethodCallTask(DBus::Connection& conn, DBus::Message& msg) : conn(conn), msg(msg) {}
 
 	DBus::Connection& conn;
 	DBus::Message msg;
     };
 
-    boost::condition_variable condition;
-    boost::mutex mutex;
-    boost::thread thread;
-    queue<Task> tasks;
+    boost::condition_variable method_call_condition;
+    boost::mutex method_call_mutex;
+    boost::thread method_call_thread;
+    queue<MethodCallTask> method_call_tasks;
+    void add_method_call_task(DBus::Connection& conn, DBus::Message& msg);
 
-    bool zombie;
+    boost::condition_variable files_transfer_condition;
+    boost::mutex files_transfer_mutex;
+    boost::thread files_transfer_thread;
+    queue<shared_ptr<FilesTransferTask>> files_transfer_tasks;
+    void add_files_transfer_task(shared_ptr<FilesTransferTask> files_transfer_task);
 
-    void add_task(DBus::Connection& conn, DBus::Message& msg);
+    bool zombie = false;
+
+    Plugins::Report report;
 
 private:
 
-    void worker();
+    void method_call_worker();
+    void files_transfer_worker();
 
     const Clients& clients;
 
@@ -173,6 +192,7 @@ private:
 class Clients
 {
 public:
+
     Clients(Backgrounds& backgrounds);
 
     typedef list<Client>::iterator iterator;
@@ -188,7 +208,8 @@ public:
 
     iterator find(const string& name);
 
-    iterator add(const string& name);
+    iterator add(const string& name, uid_t uid);
+
     void remove_zombies();
 
     bool has_zombies() const;
