@@ -1,5 +1,6 @@
 /*
  * Copyright (c) [2011-2014] Novell, Inc.
+ * Copyright (c) [2018-2024] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -22,21 +23,19 @@
 
 #include "config.h"
 
-#include <string.h>
+#include <cstring>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/mount.h>
 #include <sys/xattr.h>
+#include <sys/statvfs.h>
 #include <fcntl.h>
 #include <stddef.h>
 #include <dirent.h>
 #include <unistd.h>
-#include <errno.h>
-#include <stdlib.h>
-#include <assert.h>
-#ifdef ENABLE_SELINUX
-#include <selinux/selinux.h>
-#endif
+#include <cerrno>
+#include <cstdlib>
+#include <cassert>
 #include <algorithm>
 
 #include "snapper/FileUtils.h"
@@ -84,8 +83,11 @@ namespace snapper
 
 	dirfd = ::openat(dir.dirfd, name.c_str(), O_RDONLY | O_NOFOLLOW | O_NOATIME | O_CLOEXEC);
 	if (dirfd < 0)
-	    SN_THROW(IOErrorException(sformat("open failed path:%s errno:%d (%s)", dir.fullname().c_str(),
+	{
+	    string failed_path = dir.fullname() + "/" + name;
+	    SN_THROW(IOErrorException(sformat("open failed path:%s errno:%d (%s)", failed_path.c_str(),
 					      errno, stringerror(errno).c_str())));
+	}
 
 	struct stat buf;
 	if (fstat(dirfd, &buf) != 0)
@@ -195,6 +197,24 @@ namespace snapper
 
 	vector<string> ret;
 
+#if defined(__GLIBC__) && ((__GLIBC__ > 2) || ((__GLIBC__ == 2) && (__GLIBC_MINOR__ >= 24)))
+
+	// Since glibc 2.24 readdir is thread safe under certain
+	// conditions, which apply here, and readdir_r is deprecated
+	// (see readdir(3)).
+
+	struct dirent* ep = nullptr;
+
+	rewinddir(dp);
+	while ((ep = readdir(dp)) != nullptr)
+	{
+	    if (strcmp(ep->d_name, ".") != 0 && strcmp(ep->d_name, "..") != 0 &&
+		pred(ep->d_type, ep->d_name))
+		ret.push_back(ep->d_name);
+	}
+
+#else
+
 	long sz = fpathconf(dirfd, _PC_NAME_MAX);
 	if (sz == -1)
 	    sz = NAME_MAX;
@@ -212,6 +232,8 @@ namespace snapper
 
 	free(ep);
 
+#endif
+
 	closedir(dp);
 
 	return ret;
@@ -355,29 +377,59 @@ namespace snapper
     }
 
 
+    int
+    SDir::fsync() const
+    {
+	return ::fsync(dirfd);
+    }
+
+
+    std::pair<unsigned long long, unsigned long long>
+    SDir::statvfs() const
+    {
+	struct statvfs fsbuf;
+
+	static_assert(sizeof(fsbuf.f_blocks) == 8);
+	static_assert(sizeof(fsbuf.f_bavail) == 8);
+
+	if (fstatvfs(dirfd, &fsbuf) != 0)
+	    SN_THROW(IOErrorException(sformat("statvfs failed path:%s errno:%d (%s)", base_path.c_str(),
+					      errno, stringerror(errno).c_str())));
+
+	// f_bavail is used (not f_bfree) since df seems to do the
+	// same. Thus it allows the user to check the result easily.
+
+	return make_pair(fsbuf.f_blocks * fsbuf.f_bsize, fsbuf.f_bavail * fsbuf.f_bsize);
+    }
+
+
     int
     SDir::mktemp(string& name) const
     {
 	static const char letters[] = "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 	    "0123456789";
 
+	const size_t num_letters = strlen(letters);
+
 	static uint64_t value;
 
 	struct timeval tv;
 	gettimeofday(&tv, NULL);
 	value += ((uint64_t) tv.tv_usec << 16) ^ tv.tv_sec;
 
-	unsigned int attempts = 62 * 62 * 62;
+	unsigned int attempts = num_letters * num_letters * num_letters;
 
 	string::size_type length = name.size();
 
+	assert(length >= 6);
+
 	for (unsigned int count = 0; count < attempts; value += 7777, ++count)
 	{
 	    uint64_t v = value;
 	    for (string::size_type i = length - 6; i < length; ++i)
 	    {
-		name[i] = letters[v % 62];
-		v /= 62;
+		name[i] = letters[v % num_letters];
+		v /= num_letters;
 	    }
 
 	    int fd = open(name, O_RDWR | O_CREAT | O_EXCL | O_CLOEXEC, S_IRUSR | S_IWUSR);
@@ -574,7 +626,7 @@ namespace snapper
 
 
     bool
-    SDir::fsetfilecon(const string& name, char* con) const
+    SDir::fsetfilecon(const string& name, const char* con) const
     {
 	assert(name.find('/') == string::npos);
 	assert(name != "..");
@@ -586,8 +638,7 @@ namespace snapper
 	{
 	    char *src_con = NULL;
 
-	    int fd = ::openat(dirfd, name.c_str(), O_RDONLY | O_NOFOLLOW | O_NOATIME
-			      | O_NONBLOCK | O_CLOEXEC);
+	    int fd = ::openat(dirfd, name.c_str(), O_RDONLY | O_NOFOLLOW | O_NOATIME | O_NONBLOCK | O_CLOEXEC);
 	    if (fd < 0)
 	    {
 		// symlink, detached dev node?
@@ -610,7 +661,8 @@ namespace snapper
 		    y2deb("setting new SELinux context on " << fullname() << "/" << name);
 		    if (lsetfilecon(name.c_str(), con))
 		    {
-			y2err("lsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno << " (" << stringerror(errno) << ")");
+			y2err("lsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno <<
+			      " (" << stringerror(errno) << ")");
 			retval = false;
 		    }
 		}
@@ -625,7 +677,8 @@ namespace snapper
 		    y2deb("setting new SELinux context on " << fullname() << "/" << name);
 		    if (::fsetfilecon(fd, con))
 		    {
-			y2err("fsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno << " (" << stringerror(errno) << ")");
+			y2err("fsetfilecon on " << fullname() << "/" << name << " failed errno: " << errno <<
+			      " (" << stringerror(errno) << ")");
 			retval = false;
 		    }
 		}
@@ -636,48 +689,48 @@ namespace snapper
 	    freecon(src_con);
 	}
 #endif
+
 	return retval;
     }
 
 
     bool
-    SDir::restorecon(const string& name, SelinuxLabelHandle* sh) const
+    SDir::restorecon(const string& name, SelinuxLabelHandle* selabel_handle) const
     {
 	assert(name.find('/') == string::npos);
 	assert(name != "..");
 
 	bool retval = true;
+
 #ifdef ENABLE_SELINUX
-	if (_is_selinux_enabled())
+	assert(selabel_handle);
+
+	struct stat buf;
+	if (stat(name, &buf, AT_SYMLINK_NOFOLLOW))
 	{
-	    assert(sh);
-
-	    struct stat buf;
-	    if (stat(name, &buf, AT_SYMLINK_NOFOLLOW))
-	    {
-		y2err("Failed to stat " << fullname() << "/" << name);
-		return false;
-	    }
-
-	    char* con = sh->selabel_lookup(fullname() + "/" + name, buf.st_mode);
-	    if (con)
-	    {
-		retval = fsetfilecon(name, con);
-	    }
-	    else
-	    {
-		retval = false;
-	    }
-
-	    freecon(con);
+	    y2err("Failed to stat " << fullname() << "/" << name);
+	    return false;
 	}
+
+	char* con = selabel_handle->selabel_lookup(fullname() + "/" + name, buf.st_mode);
+	if (con)
+	{
+	    retval = fsetfilecon(name, con);
+	}
+	else
+	{
+	    retval = false;
+	}
+
+	freecon(con);
 #endif
+
 	return retval;
     }
 
 
     bool
-    SDir::fsetfilecon(char* con) const
+    SDir::fsetfilecon(const char* con) const
     {
 	bool retval = true;
 
@@ -699,41 +752,41 @@ namespace snapper
 	    freecon(src_con);
 	}
 #endif
+
 	return retval;
     }
 
 
     bool
-    SDir::restorecon(SelinuxLabelHandle* sh) const
+    SDir::restorecon(SelinuxLabelHandle* selabel_handle) const
     {
 	bool retval = true;
+
 #ifdef ENABLE_SELINUX
-	if (_is_selinux_enabled())
+	assert(selabel_handle);
+
+	struct stat buf;
+
+	if (stat(&buf))
 	{
-	    assert(sh);
-
-	    struct stat buf;
-
-	    if (stat(&buf))
-	    {
-		y2err("Failed to stat " << fullname());
-		return false;
-	    }
-
-	    char* con = sh->selabel_lookup(fullname(), buf.st_mode);
-	    if (con)
-	    {
-		retval = fsetfilecon(con);
-	    }
-	    else
-	    {
-		y2war("can't get proper label for path:" << fullname());
-		retval = false;
-	    }
-
-	    freecon(con);
+	    y2err("Failed to stat " << fullname());
+	    return false;
 	}
+
+	char* con = selabel_handle->selabel_lookup(fullname(), buf.st_mode);
+	if (con)
+	{
+	    retval = fsetfilecon(con);
+	}
+	else
+	{
+	    y2war("can't get proper label for path:" << fullname());
+	    retval = false;
+	}
+
+	freecon(con);
 #endif
+
 	return retval;
     }
 
@@ -803,15 +856,16 @@ namespace snapper
 
 
     void
-    SFile::fsetfilecon(char* con) const
+    SFile::fsetfilecon(const char* con) const
     {
 	dir.fsetfilecon(name, con);
     }
 
+
     void
-    SFile::restorecon(SelinuxLabelHandle* sh) const
+    SFile::restorecon(SelinuxLabelHandle* selabel_handle) const
     {
-	dir.restorecon(name, sh);
+	dir.restorecon(name, selabel_handle);
     }
 
 
@@ -819,7 +873,8 @@ namespace snapper
 	: base_dir(base_dir), name(name_template)
     {
 	if (!base_dir.mkdtemp(name))
-	    throw runtime_error_with_errno("mkdtemp failed", errno);
+	    SN_THROW(IOErrorException(sformat("mkdtmp failed errno:%d (%s)", errno,
+					      stringerror(errno).c_str())));
     }
 
 
@@ -844,7 +899,8 @@ namespace snapper
     {
 	SDir subdir(base_dir, name);
 	if (!subdir.mount(device, mount_type, mount_flags, mount_data))
-	    throw runtime_error_with_errno("mount failed", errno);
+	    SN_THROW(IOErrorException(sformat("mount failed errno:%d (%s)", errno,
+					      stringerror(errno).c_str())));
     }
 
 
