@@ -1,6 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
- * Copyright (c) [2016,2018] SUSE LLC
+ * Copyright (c) [2016-2023] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -26,10 +26,9 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <fcntl.h>
-#include <fnmatch.h>
-#include <dirent.h>
-#include <errno.h>
-#include <string.h>
+#include <cerrno>
+#include <cstring>
+#include <regex>
 #include <boost/algorithm/string.hpp>
 
 #include "snapper/Snapshot.h"
@@ -37,18 +36,21 @@
 #include "snapper/AppUtil.h"
 #include "snapper/XmlFile.h"
 #include "snapper/Filesystem.h"
+#ifdef ENABLE_BTRFS
 #include "snapper/Btrfs.h"
+#endif
 #include "snapper/Enum.h"
 #include "snapper/SnapperTmpl.h"
 #include "snapper/SnapperDefines.h"
 #include "snapper/Exception.h"
-#include "snapper/Regex.h"
-#include "snapper/Hooks.h"
+#include "snapper/PluginsImpl.h"
+#include "snapper/ComparisonImpl.h"
 
 
 namespace snapper
 {
     using std::list;
+    using std::regex;
 
 
     std::ostream& operator<<(std::ostream& s, const Snapshot& snapshot)
@@ -60,6 +62,9 @@ namespace snapper
 
 	s << " date:\"" << datetime(snapshot.date, true, true) << "\"";
 
+	if (snapshot.read_only)
+	    s << " read-only";
+
 	if (snapshot.uid != 0)
 	    s << "uid:" << snapshot.uid;
 
@@ -77,15 +82,12 @@ namespace snapper
 
 
     Snapshot::Snapshot(const Snapper* snapper, SnapshotType type, unsigned int num, time_t date)
-	: snapper(snapper), type(type), num(num), date(date), uid(0), pre_num(0),
-	  mount_checked(false), mount_user_request(false), mount_use_count(0)
+	: snapper(snapper), type(type), num(num), date(date)
     {
     }
 
 
-    Snapshot::~Snapshot()
-    {
-    }
+    Snapshot::~Snapshot() = default;
 
 
     // Directory containing the actual content of the snapshot.
@@ -127,10 +129,25 @@ namespace snapper
     bool
     Snapshot::isReadOnly() const
     {
-	if (isCurrent())
-	    return false;
+	return read_only;
+    }
 
-	return snapper->getFilesystem()->isSnapshotReadOnly(num);
+
+    void
+    Snapshot::setReadOnly(bool read_only)
+    {
+	if (isCurrent())
+	    SN_THROW(IllegalSnapshotException());
+
+	if (Snapshot::read_only == read_only)
+	    return;
+
+	Snapshot::read_only = read_only;
+
+	snapper->getFilesystem()->setSnapshotReadOnly(num, read_only);
+
+	if (!read_only)
+	    deleteFilelists();
     }
 
 
@@ -144,7 +161,15 @@ namespace snapper
     void
     Snapshot::setDefault() const
     {
-	return snapper->getFilesystem()->setDefault(num);
+	Plugins::Report report;
+	setDefault(report);
+    }
+
+
+    void
+    Snapshot::setDefault(Plugins::Report& report) const
+    {
+	snapper->getFilesystem()->setDefault(num, report);
     }
 
 
@@ -182,27 +207,63 @@ namespace snapper
     }
 
 
+    void
+    Snapshot::deleteFilelists() const
+    {
+	SDir info_dir = openInfoDir();
+
+	// remove all filelists in the info directory of this snapshot
+	for (const string& name : info_dir.entries(is_filelist_file))
+	{
+	    info_dir.unlink(name, 0);
+	}
+
+	// remove all filelists of the snapshot in the info directories of other snapshots
+	for (const Snapshot& snapshot : snapper->getSnapshots())
+	{
+	    if (snapshot.isCurrent())
+		continue;
+
+	    SDir tmp = snapshot.openInfoDir();
+	    string name = filelist_name(snapshot.getNum());
+	    tmp.unlink(name, 0);
+	    tmp.unlink(name + ".gz", 0);
+	}
+    }
+
+
+    Snapshots::Snapshots(const Snapper* snapper)
+	: snapper(snapper)
+    {
+    }
+
+
+    Snapshots::~Snapshots() = default;
+
+
     void
     Snapshots::read()
     {
-	Regex rx("^[0-9]+$");
+	static const regex rx("[0-9]+", regex::extended);
 
 	SDir infos_dir = snapper->openInfosDir();
 
 	vector<string> infos = infos_dir.entries();
 	for (vector<string>::const_iterator it1 = infos.begin(); it1 != infos.end(); ++it1)
 	{
-	    if (!rx.match(*it1))
+	    if (!regex_match(*it1, rx))
 		continue;
 
 	    try
 	    {
 		SDir info_dir(infos_dir, *it1);
 		int fd = info_dir.open("info.xml", O_NOFOLLOW | O_CLOEXEC);
+		if (fd < 0)
+		    SN_THROW(IOErrorException("open info.xml failed"));
+
 		XmlFile file(fd, "");
 
-		const xmlNode* root = file.getRootElement();
-		const xmlNode* node = getChildNode(root, "snapshot");
+		const xmlNode* node = file.getRootElement();
 
 		string tmp;
 
@@ -244,8 +305,8 @@ namespace snapper
 
 		getChildValue(node, "cleanup", snapshot.cleanup);
 
-		const list<const xmlNode*> l = getChildNodes(node, "userdata");
-		for (list<const xmlNode*>::const_iterator it2 = l.begin(); it2 != l.end(); ++it2)
+		const vector<const xmlNode*> l = getChildNodes(node, "userdata");
+		for (vector<const xmlNode*>::const_iterator it2 = l.begin(); it2 != l.end(); ++it2)
 		{
 		    string key, value;
 		    getChildValue(*it2, "key", key);
@@ -254,16 +315,22 @@ namespace snapper
 			snapshot.userdata[key] = value;
 		}
 
-		if (!snapper->getFilesystem()->checkSnapshot(snapshot.num))
+		const Filesystem* filesystem = snapper->getFilesystem();
+
+		if (!filesystem->checkSnapshot(snapshot.num))
 		{
 		    y2err("snapshot check failed. not adding snapshot " << *it1);
 		    continue;
 		}
 
+		snapshot.read_only = filesystem->isSnapshotReadOnly(snapshot.num);
+
 		entries.push_back(snapshot);
 	    }
-	    catch (const IOErrorException& e)
+	    catch (const Exception& e)
 	    {
+		SN_CAUGHT(e);
+
 		y2err("loading " << *it1 << " failed");
 	    }
 	}
@@ -351,6 +418,7 @@ namespace snapper
 	entries.clear();
 
 	Snapshot snapshot(snapper, SINGLE, 0, (time_t)(-1));
+	snapshot.read_only = false;
 	snapshot.description = "current";
 	entries.push_back(snapshot);
 
@@ -360,6 +428,8 @@ namespace snapper
 	}
 	catch (const IOErrorException& e)
 	{
+	    SN_CAUGHT(e);
+
 	    y2err("reading failed");
 	}
 
@@ -486,12 +556,34 @@ namespace snapper
 
 	SDir info_dir = openInfoDir();
 
-	xml.save(info_dir.mktemp(tmp_name));
+	int fd = info_dir.mktemp(tmp_name);
+	if (fd < 0)
+	    SN_THROW(IOErrorException(sformat("SDir::mktemp failed, errno:%d (%s)", errno,
+					      stringerror(errno).c_str())));
+
+	fchmod(fd, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	try
+	{
+	    xml.save(fd);
+	}
+	catch (const Exception& e)
+	{
+	    SN_CAUGHT(e);
+
+	    info_dir.unlink(tmp_name, 0);
+
+	    SN_RETHROW(e);
+	}
 
 	if (info_dir.rename(tmp_name, file_name) != 0)
+	{
 	    SN_THROW(IOErrorException(sformat("rename info.xml failed infoDir:%s errno:%d (%s)",
 					      info_dir.fullname().c_str(), errno,
 					      stringerror(errno).c_str())));
+	}
+
+	info_dir.fsync();
     }
 
 
@@ -573,76 +665,77 @@ namespace snapper
     void
     Snapshot::deleteFilesystemSnapshot() const
     {
-	if (isCurrent())
-	    SN_THROW(IllegalSnapshotException());
-
 	snapper->getFilesystem()->umountSnapshot(num);
 	snapper->getFilesystem()->deleteSnapshot(num);
     }
 
 
     Snapshots::iterator
-    Snapshots::createSingleSnapshot(const SCD& scd)
+    Snapshots::createSingleSnapshot(const SCD& scd, Plugins::Report& report)
     {
 	checkUserdata(scd.userdata);
 
 	Snapshot snapshot(snapper, SINGLE, nextNumber(), time(NULL));
 	snapshot.uid = scd.uid;
+	snapshot.read_only = scd.read_only;
 	snapshot.description = scd.description;
 	snapshot.cleanup = scd.cleanup;
 	snapshot.userdata = scd.userdata;
 
-	return createHelper(snapshot, getSnapshotCurrent(), scd.read_only, scd.empty);
+	return createHelper(snapshot, getSnapshotCurrent(), scd.empty, report);
     }
 
 
     Snapshots::iterator
-    Snapshots::createSingleSnapshot(const_iterator parent, const SCD& scd)
+    Snapshots::createSingleSnapshot(const_iterator parent, const SCD& scd, Plugins::Report& report)
     {
 	checkUserdata(scd.userdata);
 
 	Snapshot snapshot(snapper, SINGLE, nextNumber(), time(NULL));
 	snapshot.uid = scd.uid;
+	snapshot.read_only = scd.read_only;
 	snapshot.description = scd.description;
 	snapshot.cleanup = scd.cleanup;
 	snapshot.userdata = scd.userdata;
 
-	return createHelper(snapshot, parent, scd.read_only);
+	return createHelper(snapshot, parent, scd.empty, report);
     }
 
 
     Snapshots::iterator
-    Snapshots::createSingleSnapshotOfDefault(const SCD& scd)
+    Snapshots::createSingleSnapshotOfDefault(const SCD& scd, Plugins::Report& report)
     {
 	checkUserdata(scd.userdata);
 
 	Snapshot snapshot(snapper, SINGLE, nextNumber(), time(NULL));
 	snapshot.uid = scd.uid;
+	snapshot.read_only = scd.read_only;
 	snapshot.description = scd.description;
 	snapshot.cleanup = scd.cleanup;
 	snapshot.userdata = scd.userdata;
 
-	return createHelper(snapshot, end(), scd.read_only);
+	return createHelper(snapshot, end(), false, report);
     }
 
 
     Snapshots::iterator
-    Snapshots::createPreSnapshot(const SCD& scd)
+    Snapshots::createPreSnapshot(const SCD& scd, Plugins::Report& report)
     {
 	checkUserdata(scd.userdata);
 
 	Snapshot snapshot(snapper, PRE, nextNumber(), time(NULL));
 	snapshot.uid = scd.uid;
+	snapshot.read_only = scd.read_only;
 	snapshot.description = scd.description;
 	snapshot.cleanup = scd.cleanup;
 	snapshot.userdata = scd.userdata;
 
-	return createHelper(snapshot, getSnapshotCurrent(), scd.read_only);
+	return createHelper(snapshot, getSnapshotCurrent(), false, report);
     }
 
 
     Snapshots::iterator
-    Snapshots::createPostSnapshot(Snapshots::const_iterator pre, const SCD& scd)
+    Snapshots::createPostSnapshot(Snapshots::const_iterator pre, const SCD& scd, Plugins::Report& report)
     {
 	if (pre == entries.end() || pre->isCurrent() || pre->getType() != PRE ||
 	    findPost(pre) != entries.end())
@@ -653,27 +746,30 @@ namespace snapper
 	Snapshot snapshot(snapper, POST, nextNumber(), time(NULL));
 	snapshot.pre_num = pre->getNum();
 	snapshot.uid = scd.uid;
+	snapshot.read_only = scd.read_only;
 	snapshot.description = scd.description;
 	snapshot.cleanup = scd.cleanup;
 	snapshot.userdata = scd.userdata;
 
-	return createHelper(snapshot, getSnapshotCurrent(), scd.read_only);
+	return createHelper(snapshot, getSnapshotCurrent(), false, report);
     }
 
 
     Snapshots::iterator
-    Snapshots::createHelper(Snapshot& snapshot, const_iterator parent, bool read_only,
-			    bool empty)
+    Snapshots::createHelper(Snapshot& snapshot, const_iterator parent, bool empty, Plugins::Report& report)
     {
 	// parent == end indicates the btrfs default subvolume. Unclean, but
 	// adding a special snapshot like current needs too many API changes.
 
+	Plugins::create_snapshot(Plugins::Stage::PRE_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 snapshot, report);
+
 	try
 	{
 	    if (parent != end())
-		snapshot.createFilesystemSnapshot(parent->getNum(), read_only, empty);
+		snapshot.createFilesystemSnapshot(parent->getNum(), snapshot.read_only, empty);
 	    else
-		snapshot.createFilesystemSnapshotOfDefault(read_only);
+		snapshot.createFilesystemSnapshotOfDefault(snapshot.read_only);
 	}
 	catch (const CreateSnapshotFailedException& e)
 	{
@@ -700,70 +796,85 @@ namespace snapper
 	    SN_RETHROW(e);
 	}
 
-	Hooks::create_snapshot(snapper->subvolumeDir(), snapper->getFilesystem());
+	Plugins::create_snapshot(Plugins::Stage::POST_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 snapshot, report);
 
 	return entries.insert(entries.end(), snapshot);
     }
 
 
-    static bool
-    is_filelist_file(unsigned char type, const char* name)
-    {
-	return (type == DT_UNKNOWN || type == DT_REG) && (fnmatch("filelist-*.txt", name, 0) == 0);
-    }
-
-
     void
-    Snapshots::modifySnapshot(iterator snapshot, const SMD& smd)
+    Snapshots::modifySnapshot(iterator snapshot, const SMD& smd, Plugins::Report& report)
     {
 	if (snapshot == entries.end() || snapshot->isCurrent())
 	    SN_THROW(IllegalSnapshotException());
 
 	checkUserdata(smd.userdata);
 
+	Plugins::modify_snapshot(Plugins::Stage::PRE_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 *snapshot, report);
+
 	snapshot->description = smd.description;
 	snapshot->cleanup = smd.cleanup;
 	snapshot->userdata = smd.userdata;
 
 	snapshot->writeInfo();
 
-	Hooks::modify_snapshot(snapper->subvolumeDir(), snapper->getFilesystem());
+	Plugins::modify_snapshot(Plugins::Stage::POST_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 *snapshot, report);
     }
 
 
     void
-    Snapshots::deleteSnapshot(iterator snapshot)
+    Snapshots::deleteSnapshot(iterator snapshot, Plugins::Report& report)
     {
-	if (snapshot == entries.end() || snapshot->isCurrent())
+	if (snapshot == entries.end() || snapshot->isCurrent() || snapshot->isDefault() ||
+	    snapshot->isActive())
 	    SN_THROW(IllegalSnapshotException());
 
+	Plugins::delete_snapshot(Plugins::Stage::PRE_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 *snapshot, report);
+
 	snapshot->deleteFilesystemSnapshot();
+	snapshot->deleteFilelists();
 
 	SDir info_dir = snapshot->openInfoDir();
-
 	info_dir.unlink("info.xml", 0);
 
-	vector<string> tmp1 = info_dir.entries(is_filelist_file);
-	for (vector<string>::const_iterator it = tmp1.begin(); it != tmp1.end(); ++it)
-	{
-	    info_dir.unlink(*it, 0);
-	}
-
-	for (Snapshots::iterator it = begin(); it != end(); ++it)
-	{
-	    if (!it->isCurrent())
-	    {
-		SDir tmp2 = it->openInfoDir();
-		tmp2.unlink("filelist-" + decString(snapshot->getNum()) + ".txt", 0);
-	    }
-	}
-
 	SDir infos_dir = snapper->openInfosDir();
 	infos_dir.unlink(decString(snapshot->getNum()), AT_REMOVEDIR);
 
-	entries.erase(snapshot);
+	Plugins::delete_snapshot(Plugins::Stage::POST_ACTION, snapper->subvolumeDir(), snapper->getFilesystem(),
+				 *snapshot, report);
 
-	Hooks::delete_snapshot(snapper->subvolumeDir(), snapper->getFilesystem());
+	entries.erase(snapshot);
+    }
+
+
+    Snapshots::iterator
+    Snapshots::getDefault()
+    {
+	std::pair<bool, unsigned int> tmp = snapper->getFilesystem()->getDefault();
+
+	return tmp.first ? find(tmp.second) : end();
+    }
+
+
+    Snapshots::const_iterator
+    Snapshots::getDefault() const
+    {
+	std::pair<bool, unsigned int> tmp = snapper->getFilesystem()->getDefault();
+
+	return tmp.first ? find(tmp.second) : end();
+    }
+
+
+    Snapshots::const_iterator
+    Snapshots::getActive() const
+    {
+	std::pair<bool, unsigned int> tmp = snapper->getFilesystem()->getActive();
+
+	return tmp.first ? find(tmp.second) : end();
     }
 
 
